#!/usr/bin/perl
#
# Copyright (C) 2023 Juan Domingo (Juan.Domingo@uv.es)
#
# This script is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

use strict;
use warnings;
use Scalar::Util qw(looks_like_number);
use List::MoreUtils qw(uniq);

if ($#ARGV != 1)
{
 print"Usage: $0 csv_file sql_file\n";
 exit(1);
}

# Fill these variables with the location of the appropiate files:
my $GeneSQLFile    = "../FillGene.sql";
my $PatSQLFile     = "../FillPathology.sql";
my $DrugSQLFile    = "../FillDrug.sql";
my $TherapySQLFile = "../FillTherapy.sql";
my $VarTypeFile    = "../FillVariationType.sql";

my $CurrentGFP=0;
my $CurrentIdS=0;
my $CurrentIdC=0;

# The variable Gene will contain all the known genes with its EntrezID and name (as a hash). It is filled using the previously created gene database.
# and it is here to check that the gene Entrez that the user mentions really exist (and to check which one it is).
my %Gene;

# The variable Pat will contain all the known pathologies with their ATC codes (as a hash). It is filled using the previously created ICD10 database.
my %Pat;

# The variable Ter will contain all the known basic therapies. It is filled using the previously created Therapy database.
# Combinations are possible, too, but they are not stored here but tested later.
my %Ter;

# The variable VTypes will contain all the known variation type names. It is filled using the previously created VariationTypes database.
# Only a single value is possible at any time
my %VTypes;

# The variable Drugs will contain all the known drugs with their ATC code (as a hash). It is filled using the previously created ATC codes database.
# To be as leniant as possible in the drug name matching, not only the ATC name is stored but also that name in lower case and with sequences of
# more than one spaces and/or tabs replaced by a single whitespace. Both hashes are linked: the value of the normalized version will be the key
# of the original version, and the value of the original version will be the ATC code.
my %Drugs;
my %NormNamesDrugs;

# The variable PKNames will contain the names of the panel kits, as they are introduced. We can't fill them in advance, as in former cases, since
# there is no a know, closed list. 
my %PKNames;

# The variable AddedVariations will contain the identifiers of the variations that have been read and then, added to the database.
# This is to check if any of them could be repeated (which is OK) but in such a case should no be added to the database Variation table.
my %AddedVariations;

# The variable AddedSymptoms will contain the identifiers of the symptoms that have been read and then, added to the database. 
# This is to check if any of them could be repeated (which is OK) but in such a case should no be added to the database CardiovascularSymptom table.
my %AddedSymptoms;

# By the way: variables AddedVariations and AddedSymptoms are hashes instead of arrays becasue in Perl is easier to check for membership with 'exists'
# in a hash rather than running through a list. The key is the information and the value is irrelevant (we'll use just 1).

# Now, a hash with the maximum length that, according to the table creation rules, some values may have.
# This is not for all values but only for those that can be filled from studies, and therefore are prone to error.
my %MaxLen=(
"IdPanelKitName",45,
"Company",45,
"SampleType",45,
"DetectionMethod",45,
"ValidationMethod",45,
"DOIReference",256,
"StatisticalMethod",64,
"StatisticalSignificanceMeaning",32,
"StudyDesign",64,
"IdCardiovascularSymptom",256);

# Function NotUnknown(V)  (Auxiliary function to cope with possible unknown values written in different or misspelled ways
# Parameter:
#  V: The value to be checked
#  Returns true/1 if V is not the value "unknown"
sub NotUnknown
{
 if ( ($_[0] =~ /^(U|u)nspecified$/ ) || ($_[0] =~ /^(U|u)nknown$/) )
 {
  return 0;
 }
 return 1;
}

# Function: FillGenes (Fills the Gene hash with the list of all genes, extracted from a SQL file generated by R)
# Parameters: 
#        GFile: the SQL file to fill the Gene table
sub FillGenes
{
 my $GFile = $_[0];
 
 open(F," < $GFile")|| die("Cannot open file $GFile\n");
 my $line;
 my $nline=1;
 my $gene_with_tick;
 while (defined($line = <F>))
 {
  chomp($line);
  if ($line =~ /INSERT\ INTO\ Gene\ VALUES\(\'([^\']+)\'\,\'([^\']+)\'\)\;/)
  {
   $Gene{$1}=$2;
  }
  else
  {
   if ($line =~ /INSERT\ INTO\ Gene\ VALUES\(\'([^\']+)\'\,\'([^\']+)\\\'([^\']+)\'\)\;/)
   {
    $gene_with_tick=$2."'".$3;
    $Gene{$1}=$gene_with_tick;
   }
   else
   {
    print "WARNING: line $nline in file $GFile does not seem to have the right format.\n";
    print "line is:'$line'\n";
   }
  }
  $nline++;
 }
 close(F);
}

# Function: FillPathologies (Fills the Pat hash with the list of ICD codes of all pathologies, extracted from a SQL file generated by a Perl script from the ICD10 list).
# Parameters: 
#        PFile: the SQL file to fill the Pathology table
sub FillPat
{
 my $PFile = $_[0];

 open(F," < $PFile") || die("Cannot open file $PFile\n");
 my $line;
 my $nline=1;
 my $disease_with_tick;
 while (defined($line = <F>))
 {
  chomp($line);
  if ($line =~ /INSERT\ INTO\ Pathology\ VALUES\(\'([^\']+)\'\,[\s\t]*\'([^\']+)\'\)\;/)
  {
   $Pat{$1}=$2;
  }
  else
  {
   if ($line =~ /INSERT\ INTO\ Pathology\ VALUES\(\'([^\']+)\'\,[\s\t]*\'([^\']+)\\\'([^\']+)\'\)\;/)
   {
    my $disease_with_tick=$2."'".$3;
    $Pat{$1}=$disease_with_tick;
   }
   else
   {
    print "WARNING: line $nline in file $PFile does not seem to have the right format.\n";
    print "line is:'$line'\n";
   }
  }
  $nline++;
 }
 close(F);
}

# Function: NormDrugName (Normalizes the names of the drugs to make the matching wider
sub NormDrugName
{
 my $Dname = $_[0];
 $Dname = lc($Dname);
 $Dname =~ s/[\s\t]+/\_/g;
 return $Dname;
}

# Function: FillDrugs (Fills the Drug hash with the list of ATC codes, extracted from a SQL file generated by a Perl script from the ATC list).
# Parameters: 
#        DFile: the SQL file to fill the Drug table
sub FillDrug
{
 my $DFile = $_[0];
 
 open(F," < $DFile") || die("Cannot open file $DFile\n");
 my $line;
 my $nline=1;
 while (defined($line=<F>))
 {
  chomp($line);
  if ($line =~ /INSERT\ INTO\ Drug\ VALUES\(\'([^\']+)\'\,[\s\t]*\'([^\']+)\'\)\;/)
  {
   $Drugs{$2}=$1;
   $NormNamesDrugs{NormDrugName($2)}=$2;
   # print "Drugs{$2} --> $1;     NormNamesDrugs{".NormDrugName($2)."} --> $2\n";
  } 
  else
  {
   print "WARNING: line $nline in file $DFile does not seem to have the right format.\n";
   print "line is:'$line'\n";
  }
  $nline++;
 }
 close(F);
}

# Function: FillTherapies (Fills the Therapy hash with the list of therapies, extracted from a SQL file generated by a Perl script from basic information
# Parameters: 
#        TFile: the SQL file to fill the Therapy table
sub FillTherapies
{
 my $TFile = $_[0];
 
 open(F," < $TFile") || die("Cannot open file $TFile\n");
 my $line;
 my $nline=1;
 my $t;
  while (defined($line=<F>))
 {
  chomp($line);
  if ($line =~ /INSERT\ INTO\ Therapy\ VALUES\(\'([^\,]+)\'\)\;/)
  {
   $t=$1;
   if (!($t =~ /\+/))
   {
    $Ter{$t}=1;
   }
  }
  else
  {
   print "WARNING: line $nline in file $TFile does not seem to have the right format.\n";
   print "line is:'$line'\n";
  }
  $nline++;
 }
 close(F);
}

# Function: FillVTNames (Fills the Variation types hash with the list of variation type, extracted from a SQL file generated by a Perl script from basic information
# Parameters: 
#        VTFile: the SQL file to fill the VariationType table
sub FillVTNames
{
 my $VTFile = $_[0];
 
 open(F," < $VTFile") || die("Cannot open file $VTFile\n");
 my $line;
 my $nline=1;
 while (defined($line=<F>))
 {
  chomp($line);
  if ($line =~ /INSERT\ INTO\ VariationType\ VALUES\(\'([^\,]+)\'\)\;/)
  {
   $VTypes{$1}=1;
  }
  else
  {
   print "WARNING: line $nline in file $VTFile does not seem to have the right format.\n";
   print "line is:'$line'\n";
  }
  $nline++;
 }
 close(F);
}

# Function ATCOf(Drug)
# Parameter:
#    Drug: the name of a drug
# Returns the ATC code of the drug, or unknown if there is not such code or cannot be found.
# WARNING: some ATC names may contain comma (,) but they come here as semicolons (;).
# This is why they are converted before looking at them in the hash.
# OJO: repasar. Aquí se deberían mirar similaridades por errores en los nombres de los medicamentos o denominaciones muy parecidas.
# De momento, solo lo que hay.
sub ATCOf
{
 if (NotUnknown($_[0]))
 {
  # This is the substitution to go back to comma.
  $_[0] =~ s/\;/\,/g;
  exists($NormNamesDrugs{NormDrugName($_[0])}) || die("Cannot find the ATC code for drug named $_[0] whose normalized name is ".NormDrugName($_[0])."\n");
  return $Drugs{$NormNamesDrugs{NormDrugName($_[0])}};
 }
 else
 {
  return "unknown";
 }
}

# Function SQLWrite Add quotes to something or not so that it can written in the SQL.
# Things written 'as is' (i.e.: without quotes) are either numbers, or NULL, TRUE or FALSE written in any upper/lower case combination
# (to prevent mistakes). Nevertheless, they are normalized to their upper-case versions.
# The unusual case that the words null, true or false could be part of a word is contemplated (it must be the only word in the value)
sub SQLWrite
{
 if (looks_like_number($_[0]))
 {
  return $_[0];
 }
 if ($_[0] =~ /^[\s\t]*(N|n)(U|u)(L|l)(L|l)[\s\t]*$/)
 {
  return "NULL";
 }
 if ($_[0] =~ /^[\s\t]*(T|t)(R|r)(U|u)(E|e)[\s\t]*$/)
 {
  return "TRUE";
 }
 if ($_[0] =~ /^[\s\t]*(F|f)(A|a)(L|l)(S|s)(E|e)[\s\t]*$/)
 {
  return "FALSE";
 }
 
 # If is it not any of the former things, it will be returned surrounded by single quotes.
 return "'".$_[0]."'";
   
}

# Function ATD(G,T,Values...) (Add to database)
# Parameters:
#   G: file handle to write in
#   T: table name
#   Values: array of (key1,value1,key2,value2,...,keyn,valuen)
#
sub ATD
{
 my ($G,$tabname,@keysvalues) = @_;
 my $lim=$#keysvalues-1;
 print $G "INSERT INTO $tabname (";
 for (my $i=0; $i < $lim+1; $i+=2)
 {
  print $G $keysvalues[$i].(($i < $lim) ? "," : ") ");
 }
 print $G "VALUES (";
 for (my $i=0; $i <= $lim; $i+=2)
 {
  print $G SQLWrite($keysvalues[$i+1]).(($i < $lim) ? "," : ");\n");
 }
}

# Function AddGFP(G,GenomeVersionNameVal,StartPositionVal,EndPositionVal,IdGeneEntrezValue) (Adds a new genome full position)
# Parameters:
#  G: file handle to write in
#  The four next parameters are strings with meaning implied by their names
#  Last parameter (IdGeneEntrezValue) is not really needed, except to print the message.
sub AddGFP
{
 my ($G,$GenomeVersionNameVal,$StartPositionVal,$EndPositionVal,$IdGeneEntrezValue) = @_;
 
 # A new numeric indentifier is generated for this genome full position.
 my $NewGFP = $CurrentGFP;
 
 ATD($G,"GenomeFullPosition",
     ("IdGenomeFullPosition",$NewGFP,"GenomeVersionName",$GenomeVersionNameVal,"StartPosition",$StartPositionVal,"EndPosition",$EndPositionVal));
 
 print "   Gene:\n";
 print "      Entrez: $IdGeneEntrezValue";
 if ($IdGeneEntrezValue == -1)
 {
  print " (special arbitrary code for 'Intergenic')\n";
 }
 else
 {
  print "\n";
 }
 print "      GenomeVersionName: $GenomeVersionNameVal\n";
 print "      StartPosition: $StartPositionVal\n";
 print "      EndPosition: $EndPositionVal\n";
     
 $CurrentGFP++;
 return $NewGFP;
}

# Function AddPanelKit(G,IdPanelKitNameValue,CompanyValue,IdVariationValue) (Adds a new panel kit)
# Parameters:
#   G: file handle to write in
#   The three next parameters are strings with meaning implied by their names 
sub AddPanelKit
{
 my ($G,$IdPanelKitNameValue,$CompanyValue,$IdVariationValue) = @_;
 # Panel kits are added, except if unknown or if they had been added before
 print "   PanelKit:\n";
 if ( NotUnknown($IdPanelKitNameValue) && NotUnknown($CompanyValue) )
 {
  if (exists($PKNames{$IdPanelKitNameValue})) 
  {
   print "      Name: '$IdPanelKitNameValue'\n";
   print "      Already present. Not added to DB.\n";
  }
  else
  {
   ATD($G,"PanelKitName",("IdPanelKitName",$IdPanelKitNameValue,"Company",$CompanyValue,"IdVariation",$IdVariationValue));
   print "      Name: '$IdPanelKitNameValue'\n";
   print "      Company: $CompanyValue\n";
   print "      Variation: $IdVariationValue\n";
   print "      First appearance. Added to database.\n";
   $PKNames{$IdPanelKitNameValue}=1;
  }
 }
 else
 {
  print "      Either the name or the company are unknown. Not added to DB.\n";
 }
 return;
}

# Function AddAdditionalPub(G,DOIDrug) (Adds an additional publication to the table of publications on drugs).
# Parameters:
#   G: file handle to write in
#   The parameter is a strings with meaning implied by its name
sub AddAdditionalPub
{
 my ($G,$DOIDrugValue) = @_;
 # Publications are added, except if unknown. And yes, in this case it could be unknown... 
 print "   PubOnDrug (additional publication): ";
 if ( NotUnknown($DOIDrugValue) )
 {
  ATD($G,"PubOnDrug","DOIDrug",$DOIDrugValue);
  print "$DOIDrugValue\n";
  return;
 }
 print "none. Nothing to add to DB.\n";
 return;
}

# Function AddVariation(G,IdVarValue,NewIdG,VariationTypeNameValue,NewGFP) (Adds a variation)
# Parameters:
#   G: file handle to write in
#   The next four parameters are strings with meaning implied by their names
# WARNING:
# 	Notice that variations are introduced "on the fly", i.e.: always, as long as they are not been introduced before.
#	No attempt is done to check the validity of the variation name (even the previous analysis shows clues on it).
sub AddVariation
{
 my ($G,$IdVarValue,$NewIdG,$VariationTypeNameValue,$NewGFP) = @_;
 print "   Variation:\n";
 print "      Identifier: $IdVarValue\n";
 print "      Type: $VariationTypeNameValue\n";
 if (!exists($AddedVariations{$IdVarValue}))
 {
  ATD($G,"Variation","IdVariation",$IdVarValue,"IdGeneEntrez",$NewIdG,"VariationTypeName",$VariationTypeNameValue,"IdGenomeFullPosition",$NewGFP);
  $AddedVariations{$IdVarValue}=1;              # Just to mark it exists.
  print "      First appearance of the variation (not necessarily of the type). Added to DB.\n";
 }
 else
 {
  print "      Not added to DB since it was already present.\n";
 }
}

# Function CheckIndividualTherapy($t)
# Checks if the argument is a valid therapy name, admitting upper/lower case.
# The normalized therapy name (we decide that normalized means 'all lower case') is returned.
sub CheckIndividualTherapy
{
 my $t=lc($_[0]);
 foreach my $item(keys(%Ter))
 {
  if ($t =~ /$item/)
  {
   return $t;
  }
 }
 die("Error: $t is not a valid therapy name.\n");
}

# Function CheckTherapies($lt) (Checks if a therapy, or a list of them separated by comma, is valid (are all valid) and, if so, returns them as an array of strings
# Parameters:
#    $lt: a string of the form t or t1,t2,.. where all t must be in 'chemotherapy','radiotherapy','targeted therapy','immunotherapy', 'surgery','hormonal'
#         or any combinations of them separated by the '+' sign.
#         Note: upper/lower case of first letter is allowed.
# Returns: 
#    A list of strings with the valid names of all therapies, as strings, if they are all valid
#    If any of the therapies is not valid the program dies
sub CheckTherapies
{
 my $lt=$_[0];
 my @r=();
 
 if ( NotUnknown($lt) )
 {
  my @l=split(/,/,$lt);
  foreach my $item(@l)
  {
   # Get rid of front/end spaces
   $item =~ s/^[\s\t]+//g;
   $item =~ s/[\s\t]+$//g;
   # Again, separate the combinations of therapies, if it is the case.
   my @l2=split(/\+/,$item);
   # @r2 is an auxiliary array to get each element of the combined therapy
   my @r2=();
   foreach my $item2(@l2)
   {
    # Again, get rid of front/end spaces
    $item2 =~ s/^[\s\t]+//g;
    $item2 =~ s/[\s\t]+$//g;
    push(@r2,CheckIndividualTherapy($item2));
   }
   # All the individual therapies are alphabetically sorted (to be stored as a normalized value in the DB) and joined by the '+' sign.
   # Also, if any of them was repeated it is eliminated (by uniq).
   push(@r,join('+',uniq(sort(@r2)))); 
  }
  return uniq(@r);
 }
 else 
 {
  # The value 'unknown' is allowed for a therapy. It means the paper does not specify the used therapy (which is strange, anyway..)
  # It is not the same as the value 'none' which would mean that no therapy was applied (even more strange, but, well...)
  push(@r,"unknown");
  return @r; 
 }
}

# Function CheckPathologies($lt) (Checks if a pathology, or a list of them separated by comma, is valid (are all valid) and, if so, returns them as an array of strings
# Parameters:
#    $lt: a string of the form t or t1,t2,.. where all t must be in the ICD10 list
#         Note: upper/lower case of first letter is allowed.
# Returns: 
#    A list of strings with the valid names of all pathologies, as strings, if they are all valid
#    If any of the pathologies is not valid the program dies
sub CheckPathologies
{
 my $lt=$_[0];
 my @r=();
 
 if ( NotUnknown($lt) )
 {
  my @l=split(/,/,$lt);
  foreach my $item(@l)
  {
   # Get rid of front/end spaces
   $item =~ s/^[\s\t]+//g;
   $item =~ s/[\s\t]+$//g;
   if (exists($Pat{$item}))
   {
    push(@r,$item);
   }
   else
   {
    die("Pathology $item is not in the ICD10 list.\n");
   }
  }
  return @r;
 }
 else 
 {
  die("Value 'unknown' is not allowed for a pathology. Simply get rid of them in the list. If it was the only one (i.e.: the value of the pathology is unknow) this study would not be valid.\n");
 }
}

# Function CheckDrugs($ld) (Checks if a drug, or a list of them separated by comma, is valid (are all valid) and, if so, returns them as an array names.
# WARNING: some ATC names may contain commas, which is a problem to distinguish from the comma we use as separator between tow different drugs.
# Therefore, the input will use semmicolon (;) to indicate a comma in the drug name, and the function ATCof will know about that.
# Parameters:
#    $ld: a string of the form d or d1,d2,.. where all d should be drug names with existent ATC code
sub CheckDrugs
{
 my $ld=$_[0];
 my @r=();
 
 if ( NotUnknown($ld) )
 {
  my @l=split(',',$ld);
  foreach my $item(@l)
  {
   # Get rid of front/end spaces
   $item =~ s/^[\s\t]+//g;
   $item =~ s/[\s\t]+$//g;
   # ATCOf kills the program with an error message if this drug has no known ATC
   ATCOf($item);
   push(@r,$item);
  }
  return @r;
 }
 else
 {
  push(@r,"unknown");
  return @r;
 }
}

# Function NormalizedVTName(VTName) (Returns the normalized form of the variation type name or dies if the argument is invalid
# Parameters:
#   VTName: a string with a candidate name of variation type
# Returns:
#   The normalized name
sub NormalizedVTName
{
 my ($Vn,$line,$nl) = @_;
 
 # First, several consecutive spaces/tabs are normalized to just one space
 my $Vn2=$Vn;
 $Vn2 =~ s/[\s\t]+/\ /g;
 # then, string is converted to lower case
 $Vn2 = lc($Vn2);
 # and comparisons are done in lower case (but the returned value if successful is the stored key)
 
 foreach my $item(keys(%VTypes))
 {
  if ($Vn2 eq lc($item))
  {
   return $item;
  }
 }
 die("Error in line $nl, which is\n'$line'\n  '$Vn' (converted to '$Vn2') is not a valid name for the type of a variation.\n");
}

# Function ListOfSymptoms(Symptoms_with_comma)
# Returns the symptoms as a list
# Parameter:
#  Symptoms_with_comma: a string of comma-sepparated symptoms with meaning implied by its name
sub ListOfSymptoms
{
 my $S=shift;
 my @l=split(/,/,$S);
 my @ret;
 foreach my $item(@l)
 {
  # Get rid of front/end spaces
  $item =~ s/^[\s\t]+//g;
  $item =~ s/[\s\t]+$//g;
  push(@ret,$item);
 }
 return @ret;
}

# Function AddCardiovascularSymptom(G,Symptom) (Adds a cardiovascular symptom to the table CardiovascularSymptom).
# Parameters:
#   G: file handle to write in
#   SymptomValue: a string of comma-sepparated symptoms with meaning implied by its name
sub AddCardiovascularSymptom
{
 my ($G,$SymptomValue) = @_;

 my @l=ListOfSymptoms($SymptomValue);
 foreach my $item(@l)
 {
  # If the symptom is unknown/unspecified, obviously nothing will be added to the table.
  # Otherwise, its Id (which in this case is just the symptom's description) is added, if it was not already in the table.
 
  print "   CardiovascularSymptom:\n";
  if ( NotUnknown($item) )
  {
   print "      Symptom description: '$item'\n";
   if ( !exists($AddedSymptoms{$item}) )
   {
    ATD($G,"CardiovascularSymptom","IdCardiovascularSymptom",$item);
    $AddedSymptoms{$item}=1;              # Just to mark it exists.
    print "      First appearance. Added to DB.\n";
   }
   else
   {
    print "      Not added to DB since it was already present.\n";
   }
  }
  else
  {
   print "      'unspecified' (obviously, not added).\n";
  }
 }
}

# Function AddCohortInfo(G,MeanAgeValue,MaleProportionValue,CountryValue,StudyDesignValue,IdStudyValue)
# Parameters:
#   G: file handle to write in strings with meaning implied by their names 
# Returns:
#   The new conssecutive integer identified generated for this cohort
sub AddCohortInfo
{
 my ($G,$MeanAgeValue,$MaleProportionValue,$CountryValue,$StudyDesignValue,$IdStudyValue) = @_;
 print "   Cohort info:\n";
 
 # A new numeric identifier is generated for this cohort
 my $NewIdC = $CurrentIdC;
 
 ATD($G,"CohortInfo",
   ("IdCohortInfo",$NewIdC,
    "MeanAge",$MeanAgeValue,
    "MaleProportion",$MaleProportionValue,
    "Country",$CountryValue,
    "StudyDesign",$StudyDesignValue,
    "IdStudy",$IdStudyValue));
 
 print "      MeanAge: $MeanAgeValue\n";
 print "      MaleProportion: $MaleProportionValue\n";
 print "      Country: $CountryValue\n";
 print "      StudyDesign: $StudyDesignValue\n";
 print "      IdStudy: $IdStudyValue\n";
 
 $CurrentIdC++;
 return $NewIdC;
}

# Function AddStudy(G,MinObservedPropPopulationValue,MaxObservedPropPopulationValue,
#			ChangeObservedValue,LFCValue,StatisticalSignificanceMeaningValue,StatisticalMethodValue,
#			StatisticalSignificanceValueValue,StatisticalSignificanceSecValueValue,SampleTypeValue,DetectionMethodValue,ValidationMethodValue,
#			DetectionAgreesValidationValue,NumSubjectsValue,DOIReferenceValue,RefDateValue,
#			CardiotoxicityAppearsValue,ObservedCardiotoxicityValue)
# Parameters:
#   G: file handle to write in
#   The next 16 parameters are strings with meaning implied by their names
# Returns:
#   The new conssecutive integer identified generated for this study
sub AddStudy
{
 my ($G,$MinObservedPropPopulationValue,$MaxObservedPropPopulationValue,
		$ChangeObservedValue,$LFCValue,$StatisticalSignificanceMeaningValue,$StatisticalMethodValue,
		$StatisticalSignificanceValueValue,$StatisticalSignificanceSecValueValue,$SampleTypeValue,$DetectionMethodValue,$ValidationMethodValue,
		$DetectionAgreesValidationValue,$NumSubjectsValue,$DOIReferenceValue,$RefDateValue,
		$CardiotoxicityAppearsValue,$ObservedCardiotoxicityValue) = @_;
 
 print "   Study:\n";
 
 # A new numeric identifier is generated for this study
 my $NewIdS = $CurrentIdS;
 
 ATD($G,"Study",
   ("IdStudy",$NewIdS,
    "MinObservedPropPopulation",$MinObservedPropPopulationValue,
    "MaxObservedPropPopulation",$MaxObservedPropPopulationValue,
    "ChangeObserved",$ChangeObservedValue,
    "LFC",$LFCValue,
    "StatisticalSignificanceMeaning",$StatisticalSignificanceMeaningValue,
    "StatisticalMethod",$StatisticalMethodValue,
    "StatisticalSignificanceValue",$StatisticalSignificanceValueValue,
    "StatisticalSignificanceSecValue",$StatisticalSignificanceSecValueValue,
    
    "SampleType",$SampleTypeValue,
    "DetectionMethod",$DetectionMethodValue,
    "ValidationMethod",$ValidationMethodValue,
    "DetectionAgreesValidation",$DetectionAgreesValidationValue,
    "NumSubjects",$NumSubjectsValue,
    "DOIReference",$DOIReferenceValue,
    "RefDate",$RefDateValue,
    "CardiotoxicityAppears",$CardiotoxicityAppearsValue,
    "ObservedCardiotoxicity",$ObservedCardiotoxicityValue));

 print "      MinObservedPropPopulation: $MinObservedPropPopulationValue\n";
 print "      MaxObservedPropPopulation: $MaxObservedPropPopulationValue\n";
 print "      ChangeObserved: $ChangeObservedValue\n";
 print "      LFC: $LFCValue\n";
 print "      StatisticalSignificanceMeaning: $StatisticalSignificanceMeaningValue\n";
 print "      StatisticalMethod: $StatisticalMethodValue\n";
 print "      StatisticalSignificanceValue: $StatisticalSignificanceValueValue\n";
 print "      StatisticalSignificanceSecValue: $StatisticalSignificanceSecValueValue\n";
 print "      SampleType: $SampleTypeValue\n";
 print "      DetectionMethod: $DetectionMethodValue\n";
 print "      ValidationMethod: $ValidationMethodValue\n";
 print "      DetectionAgreesValidation: $DetectionAgreesValidationValue\n";
 print "      NumSubjects: $NumSubjectsValue\n";
 print "      DOIReference: $DOIReferenceValue (added before)\n";
 print "      RefDate: $RefDateValue\n";
 print "      CardiotoxicityAppears: $CardiotoxicityAppearsValue\n";
 print "      ObservedCardiotoxicity: $ObservedCardiotoxicityValue\n";
 
 $CurrentIdS++;
 return $NewIdS;
}


# Function GenerateSQLRecord (Generates the SQL code to add a full record)
# Parameters:
#   sn: the number of the study (just to make it known in case of an error)
#   G: file handle to write in
#   rec: pointer to the hash that contains the record in the form ${$rec}{key} --> value
sub GenerateSQLRecord
{
 my ($sn,$G,$rec) = @_;
 
 # This is just for clarity; now, h is the hash
 my %h=%$rec;
 
 # The mentioned gene must exist in the gene table. This has been checked when the .csv was procesed.
 my $IdGeneEntrezValue = $h{"IdGeneEntrez"};

 # The mentioned ICD10 code of the pathology  must exist in the pathology table. This has been checked when the .csv was processed.
 #my $CD10Id=$h{"IdPathologyCD10"};
 #print "   Pathology/pathologies ICD10 code(s):\n";
 #print "      $CD10Id ($Pat{$CD10Id})\n";
 
 # The mentioned variation type is recovered. It is one of the know ones. This has been checked when the .csv was processed.
 my $Vtname=$h{"VariationTypeName"};
        
 # Always add the genome full position. Each one has a new primary key (an integer number) that the function returns to be used later when needed.
 # Messages handled inside.
 my $NewIdG=AddGFP($G,$h{"GenomeVersionName"},$h{"StartPosition"},$h{"EndPosition"},$IdGeneEntrezValue);
 
 # Add the cardiovascular symptom. The function checks for unknown/unspecified value
 # Messages handled inside
 AddCardiovascularSymptom($G,$h{"IdCardiovascularSymptom"});
 
 # Variation is added. WARNING: No explicit check is done.
 my $IdVar=$h{"IdVariation"};
 # Messages handled inside.
 AddVariation($G,$h{"IdVariation"},$h{"IdGeneEntrez"},$Vtname,$NewIdG);
 
 # Panel kit is added. FillPanelKit already takes into account unknown values, which are not added. Messages handled inside.
 AddPanelKit($G,$h{"IdPanelKitName"},$h{"Company"},$h{"IdVariation"}) ;
 
 # Finally, the study is added. Some of its values might be unknown or NULL, but that's OK.
 # (We check if they are numerical values, that must be NULL if unknown or unspecified or erroneous in any other way, i.e.: anything which does not seem a number).
 my $MinObservedPropPopulationValue = looks_like_number($h{"MinObservedPropPopulation"}) ? $h{"MinObservedPropPopulation"} : "NULL";
 my $MaxObservedPropPopulationValue = looks_like_number($h{"MaxObservedPropPopulation"}) ? $h{"MaxObservedPropPopulation"} : "NULL";
 my $StatisticalSignificanceValueValue = looks_like_number($h{"StatisticalSignificanceValue"}) ? $h{"StatisticalSignificanceValue"} : "NULL";
 my $StatisticalSignificanceSecValueValue = looks_like_number($h{"StatisticalSignificanceSecValue"}) ? $h{"StatisticalSignificanceSecValue"} : "NULL";
 my $NumSubjectsValue = looks_like_number($h{"NumSubjects"}) ? $h{"NumSubjects"} : "NULL";
 my $ObservedCardiotoxicityValue;
 if (looks_like_number($h{"ObservedCardiotoxicity"}))
 {
  print $h{"ObservedCardiotoxicity"}."\n";
  if ($h{"CardiotoxicityAppears"} eq "FALSE")
  {
   die("Error: ObservedCardiotoxicity seems a number but CardiotoxicityAppears is FALSE");
  }
  $ObservedCardiotoxicityValue = $h{"ObservedCardiotoxicity"};
 }
 else
 {
  print $h{"ObservedCardiotoxicity"}."\n";
  $ObservedCardiotoxicityValue = "NULL";
 }
 
 # Each study has a new primary key (an integer number) that the function returns to be used later when needed.
 # Messages handled inside.
 my $NewIdS=AddStudy($G,$MinObservedPropPopulationValue,$MaxObservedPropPopulationValue,
			$h{"ChangeObserved"},$h{"LFC"},$h{"StatisticalSignificanceMeaning"},
			$h{"StatisticalMethod"},$StatisticalSignificanceValueValue,$StatisticalSignificanceSecValueValue,
			$h{"SampleType"},$h{"DetectionMethod"},$h{"ValidationMethod"},
			$h{"DetectionAgreesValidation"},$NumSubjectsValue,$h{"DOIReference"},
			$h{"RefDate"},$h{"CardiotoxicityAppears"},$ObservedCardiotoxicityValue);
	
 # Finally, the cohort info is added.
 my $NewIdC=AddCohortInfo($G,$h{"MeanAge"},$h{"MaleProportion"},$h{"Country"},$h{"StudyDesign"},$NewIdS);
 
 # Here begins the filling of the secondary tables
 # All values they need should be present or have been generated before by the functions which add rows to the primary tables.
 
 # Several rows of the patients_affected_by table must be filled if more than one pathology is mentioned.
 # The variable observed_pathologies will contain the valid names of the observed pathologies for a given study
 # with their corresponding ICD10 codes as its values.
 my @observed_pathologies=CheckPathologies($h{"IdPathologyCD10"});
 my $ICD10PathologyValue;
 foreach my $item(@observed_pathologies)
 {
  # Fill the table patients_affected_by, which is always done, since IdStudy and IdPathologyCD10 are always present and never unknown.
  ATD($G,"patients_affected_by","IdStudy",$NewIdS,"IdPathologyCD10",$item);
 }
 
 # Notice that this secondary table is filled with the variation id, even when it was before in the Variation table.
 # Nothing to check now in this table. Varition and study always exist both.
 ATD($G,"variation_studied_by","IdVariation",$h{"IdVariation"},"IdStudy",$NewIdS);
 
 # 
 # These tables are not to be filled by this .csv. It is another one that we will treat later, or in another step.
 # AddAdditionalPub already takes into account unknown values, and does not add them.
 # my $DOIAddPubValue=$h{$DOIDrug};
 # AddAdditionalPub($G,$DOIAddPubValue);
 # if ( NotUnknown($DOIAddPubValue) )
 # {
 #  $IdDrugATCValue = ATCOf($used_drugs[0]);
 #  ATD($G,"pub_about_drug","IdDrugATC",$IdDrugATCValue,"DoiReference",$h{"DOIDrug"});
 # }
  
 # Several rows of the cancer_tr_used_drug table must be filled if more than one drug was used.
 # The variable used_drugs will contain the valid names of the used drugs for a given study
 # with their corresponding ATC codes as its values.
 my @used_drugs=CheckDrugs($h{"cancer_tr_used_drug"});
 my $IdDrugATCValue;
 foreach my $item(@used_drugs)
 {
  $IdDrugATCValue = ATCOf($item);
  if (NotUnknown($IdDrugATCValue))
  {
   # Notice that this secondary table is filled as long the drug is not unknown, even when it was before in the Drug table.
   ATD($G,"cancer_tr_used_drug","IdStudy",$NewIdS,"IdDrugATC",$IdDrugATCValue);
  }
 }
 
 # Several rows of the cardiovascular symptoms must be filled if more than one symptom is present
 # The symptom and study is added to the table exhibit_symptom if the symptom is not unknown/unspecified
 my @present_symptoms=ListOfSymptoms($h{"IdCardiovascularSymptom"});
 foreach my $item(@present_symptoms)
 {
  if ( NotUnknown($item) )
  {
   ATD($G,"exhibit_symptom","IdStudy",$NewIdS,"IdCardiovascularSymptom",$item);
  }
 }
 
 # The variable used_therapies will contain the valid names of the used therapies for a given study (as a list, may be with just one element...)
 my @used_therapies=CheckTherapies($h{"cancer_tr_used_therapy"});
 
 if ( NotUnknown($used_therapies[0]) )  # The case of error was checked when loading the .csv file and it provoked a die, but unknown is not an error.
 {
  foreach my $t(@used_therapies)
  {
   ATD($G,"cancer_tr_used_therapy","IdStudy",$NewIdS,"IdTherapy",$t);
  }
 }
}

######################################################################################################################
#                                                                                                                    #
#                                                 Main program starts here                                           #
#                                                                                                                    #
######################################################################################################################

# First, fill the list of genes
FillGenes($GeneSQLFile);

# Now, fill the list of pathologies
FillPat($PatSQLFile);

# Now, fill the list of drugs. Fortunately, no one had the stupid idea of including ticks into the name of any drug...
FillDrug($DrugSQLFile);

# Now, fill the list of therapies. This is a very limited list (currently six different therapies)
FillTherapies($TherapySQLFile);

# Finally, fill the list of variation type names. This is a limited list (currently eleven different names)
FillVTNames($VarTypeFile);

# The file fields.csv contains just the names of the valid fields, one per line, surrounded by
# double quotes. They are read and stored in the fields hash.
my %fields;
open(F," < fields.csv") || die("Cannot open fields.csv");
my $line=<F>;
my $f;
my $i=2;
# Exception: the first line is not a field, is the word Study.
($line =~ /\"Study\"/) || die("First line is not \"Study\"");
while (defined($line=<F>))
{
 if ($line =~ /\"([^\"]+)\"/)
 {
  $f=$1;
  if (exists($fields{$f}))
  {
   print "Error in line $i: duplicate field.\n";
  }
  else
  {
   $fields{$f}=$1;
  }
 }
 else
 {
  print "Error in line $i: field not surrounded by double quotes.\n";
 }
 $i++; 
}
close(F);

# The number of fields is the number of lines, minus 2: one for the first "Study" line, and one for the last increment.
my $num_fields=$i-2;

# The countrycodes file contains the two-letter ISO codes of the countries. They are stored in a hash to test for validity later.
my %ISOCountryCodes;
open(F," < countrycodes") || die("Cannot open file countrycodes to read.\n");
while (defined($line=<F>))
{
 chomp($line);
 $line =~ s/^[\s\t]*([A-Z][A-Z])[\s\t]*$/$1/; 
 ($line =~ /[A-Z][A-Z]/) || die("Error in file countrycodes: $1 is not a two-upper-case-letter country code.\n");
 $ISOCountryCodes{$line}="";
}
close(F);

# Now, let's parse the full file of data. Each study will be stored as a hash and all of them will be
# grouped in an array of hashes (@studies)

print "++++++++++++++++++++++++++++++++++++++++++++++++++++\n";
print "+                                                  +\n";
print "+      First part: read and ckeck csv file.        +\n";
print "+                                                  +\n";
print "++++++++++++++++++++++++++++++++++++++++++++++++++++\n\n";
                         
my @studies;
my $num_studies=0;
my $rec;
my $v;
my $varname;
my $genename;
my @therapies_to_check;
my @pathologies_to_check;
my @drugs_to_check;

$i=1;
open(F," < $ARGV[0]") || die("Cannot open $ARGV[0]");
while (defined($line=<F>))
{
 chomp($line);
 # When the line Study is found, the data for a new study come next, up to the following Study line (if any)
 if ($line =~ /\"Study\"/)
 {
  # If there has been a previous Study, store the current one (just finished) in the array and clean the hash.
  if ($num_studies>0)
  {
   print "   Variation name is $varname, which ".(($varname eq $genename) ? "coincides with the gene name. That seems OK.\n" : "does NOT coincide with the gene name, but this may (or may not) be correct. Check it.\n");
   push(@studies,$rec);
   # The hash pointer points now to an empty hash
   $rec={};
  }
  $num_studies++;
  print "Study $num_studies (from line "; print 38*($num_studies-1)+1; print "):\n";
  $i++;
  $varname="";
  $genename="";
 }
 else
 {
  # This is the case for a normal line (a field with a value)
  if ($line =~ /\"([^\"]+)\"[\s\t]+(.+)[\s\t]*$/)
  {
   $f=$1;
   $v=$2;
   # Let's check that the field is a valid one, at least sintactically:
   if (exists($fields{$f}))
   {
    # The value may be surrounded by double quotes, or not. If it is, get rid of them.
    if ($v =~ /\"([^\"]+)\"/)
    {
     $v=$1;
    }
    # Also, it might be a numeric value with , instead of . as decimal separator. Let's correct it.
    # This is applicable to several values, namely:
    # In study:
    #  StatisticalSignificanceValue, StatisticalSignificanceSecValue, LFC, ObservedCardiotoxycity, MinObservedPropPopulation and MaxObservedPropPopulation.
    # In CohortInfo:
    #  MeanAge and MaleProportion
    # If one of these values has no decimals, nothing is done (and it doesn't need to).
    if (($f =~ /StatisticalSignificance(Sec|)Value/) || 
        ($f =~ /LFC/) ||
        ($f =~ /ObservedCardiotoxycity/) ||
        ($f =~ /(Min|Max)ObservedPropPopulation/) ||
        ($f =~ /MeanAge/) || 
        ($f =~ /MaleProportion/)
       )
    { 
     $v =~ s/([\+\-0-9]+)\,([0-9]+)/$1\.$2/;
    }
    # and finally, initial and final spaces/tabs if any must be elliminated (but not those into the string, which can exist).
    $v =~ s/^[\s\t]+//g;
    $v =~ s/[\s\t]+$//g;
     
    # Just for debugging, since we do not check variation names...
    if ($f =~ /IdVariation/)
    {
     $varname=$v;
    }
    # Now, let's check for semantic validity of particular values:
    # First: is it the Entrez id of the gene one of those recovered in the data base?
    if ($f =~ /IdGeneEntrez/)
    {
     exists($Gene{$v}) || die("Error in line $i: '$v' is not a valid value for an IdGeneEntrez.\n");
     print "   Gene with Entrez id $v, whose name is $Gene{$v}. OK.\n";
     $genename=$Gene{$v};
    }
    # Second: is/are the CD10 code/s of the pathology/pathologies mentioned a known one?
    if ($f =~ /IdPathologyCD10/)
    {
     @pathologies_to_check=CheckPathologies($v);
     # CheckPathologies calls a die if the pathology (any pathology) is not unknown, but it is not in the ICD10.
     # exists($Pat{$v}) || die("Error in line $i: '$v' is not a valid value for an ICD10 code of a pathology.\n");
     print "   Pathology/ies observed is/are: ";
     foreach my $item(@pathologies_to_check)
     {
      print "'$Pat{$item} (ICD10: $item)' ";
     }
     print "OK\n";
    }
    # Third: the name of the therapy must be one of the few known types.
    if ($f =~ /cancer\_tr\_used\_therapy/)
    {
     @therapies_to_check=CheckTherapies($v);
     # CheckTherapies calls a die if the therapy (any therapy) is not unknown, but it is not in the list of known therapies.
     print "   Applied therapy/therapies is/are: ";
     foreach my $item(@therapies_to_check)
     {
      print "'$item' ";
     }
     print "OK.\n";
    }
    # Fourth: the name of the used drug can be a list of several drugs. Have we the ATC code for all of them?
    if ($f =~ /cancer\_tr\_used\_drug/)
    {
     @drugs_to_check=CheckDrugs($v);
     # Check drugs calls a die if the code (any code, if more than one) is not unknown, but it is not in the set of known ATC codes.
     print "   Applied drug/s is/are: ";
     foreach my $item(@drugs_to_check)
     {
      print "'$item (ATC: ".ATCOf($item).")' ";
     }
     print "OK.\n";
    }
    # Fifth: the variation type must be one of those know. It is compared in lower case, but normalized to the way in which these name were stored.
    if ($f =~ /VariationTypeName/)
    {
     
     # If this is not something that can be interpreted as a variation type valid name, the function fail and the program dies.
     # If it is, it is returned in the original way in which is it stored.
     $v = NormalizedVTName($v,$line,$i);
     print "   Variation is of type '$v'. OK\n";
    }
    
    # Sixth: the format of a date must be yyyy-mm-dd
    if ($f =~ /RefDate/)
    {
     if ($v =~ /(\d{4})\-(\d{2})\-(\d{2})/)
     {
      if (($1 < 1990) || ($1 > 2030) || ($2 < 1) || ($2 > 12) || ($3 < 1) || ($3 > 31))
      {
       die("Error in line $i: '$v' is a valid date format, but it does not express a valid date. Remember: it must be yyyy-mm-dd.\n");
      }
     }
     else
     {
      die("Error in line $i: '$v' is not a valid date format. Remember: it must be yyyy-mm-dd.\n");
     }
    }
    # Seventh: boolean values must be TRUE or FALSE, in upper or lower case
    if (($f =~ /DetectionAgreesValidation/) || ($f =~ /CardiotoxicityAppears/))
    {
     (uc($v) eq "TRUE") || (uc($v) eq "FALSE") || die("Error in line $i: '$v' is not a valid boolean value (valid values are true, True, TRUE, false, False and FALSE).\n");
     $v=uc($v);  # If it is correct, convert to upper case, to be used as SQL values.
    }
    # Eighth: possible values for StatisticalSignificanceMeaning are restricted
    if ($f =~ /StatisticalSignificanceMeaning/)
    {
     (uc($v) eq "PVAL") || 
     (uc($v) eq "HAZARD RATIO") || 
     (uc($v) eq "PVALADJUSTED") ||
     (uc($v) eq "FDR") ||
     (uc($v) eq "SENSITIVITYANDSPECIFICITY") ||
     (uc($v) eq "QVALX") ||
     (uc($v) eq "NULL") ||
     die("Error in line $i: '$v' is not a valid value for StatisticalSignificanceMeaning (valid values are pVal, Hazard Ratio, SensitivityAndSpecificity, pValAdjusted, qValx or NULL (if no ststistical method was used).\n");
    }
    # Nineth: the country code must be a two-letter ISO country code
    {
     if ($f =~ /Country/)
     {
      exists($ISOCountryCodes{uc($v)}) || die("Error in line $i: '$v' is not a valid two-letter ISO country code.\n");
      # Our convention is to store the country codes in upper-case.
      $v=uc($v); 
     }
    }
    # Finally, some values are limited in length. Let's check they do not overfit:
    foreach my $k(keys(%MaxLen))
    {
     if (($f =~ /$k/) && (length($v)>$MaxLen{$k}))
     {
      die("Error in line $i: field $f has a maximum lenght of $MaxLen{$k}. Value '$v' is longer than that.\n");
     }
    }
    
    # Now, we can store it in the hash (but we have a hash pointer, so the notation ->)
    $rec->{$f}=$v;
   }
   else
   {
    print "Error in line $i: the first field ($f) is not a valid one.\n";
    exit(1);
   }
  }
  else
  {
   print "Error in line $i, which is '$line': no two fields surrounded by double quotes.\n";
   exit(1);
  }
  $i++;
 }
}
close(F);

# Don't forget to push the last study (we didn't do the normal way, since there is no Study line to start a new one).
print "   Variation name is $varname, which ".(($varname eq $genename) ? "coincides with the gene name. That seems OK.\n" : "does NOT coincide with the gene name, but this may (or may not) be correct. Check it.\n");
push(@studies,$rec);

# To compensate for the last line increment...
$i--;

(($i % $num_studies)==0) || die("Error: the number of lines is not an exact multiple of the number of studies.\n");


# We multiply by $num_fields + 1 to take into account the Study line...
my $nf1=$num_fields+1;
print "----------------------------------------------------------------------\n";
print "     Found $num_studies studies with $nf1 fields each in a file of $i lines.";

($num_studies*$nf1==$i) || die("Error: the number of fields per study (which is $nf1, counting with the 'Study' header) multiplied by the number of studies (which is $num_studies) does not equal the number of lines.\n");

print " OK.\n";
print "----------------------------------------------------------------------\n";

print "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";
print "+                                                                   +\n";
print "+      Second part: generation of the insertion SQL scripts.        +\n";
print "+                                                                   +\n";
print "+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n";

# Finally, generate the database commands in a .sql script.
open(H,"> $ARGV[1]") || die("Cannot open file $ARGV[1] to write in it.\n");

$i=0;
my $nk;
for $rec(@studies)
{
 print "\nStudy $i with associated file study$i.sql:\n";
 $nk=scalar(keys(%$rec));
 if ($nk != $num_fields)
 {
  die("Error in study $i: it has $nk fields instead of $num_fields.\n");
 }
 open(my $G,"> study$i.sql");
 GenerateSQLRecord($i,$G,$rec);
 close($G);
 print H "select 'Loading study number $i...' as '';\n";
 print H "source study$i.sql;\n";
 $i++;
}

close(H);
print "----------------------------------------------------------------\n";
print "             End. Global script $ARGV[1] generated.\n";
print "----------------------------------------------------------------\n";

open(FS,">card_sympt.txt");
foreach my $item(sort(keys(%AddedSymptoms)))
{
 print FS "$item\n";
}
close(FS);
 
